/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Copyright 2004-2005 Simon Newton
 *  nomis52@westnet.com.au
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/wait.h>

#include <artnet/artnet.h>
#include <artnet/packets.h>

#define DEFAULT_DEVICE "/dev/dmx0"
#define UNCONNECTED -1 
#define SHORT_NAME "NWAN"
#define LONG_NAME "Netgear Wireless Artnet Node"
#define ALARM_INT 2

typedef struct {
	int verbose ;
	char *dev;
	char *ip_addr ;
	int subnet_addr ;
	int port_addr ;
	int persist ;

	int fd ;
	char *short_name ;
	char *long_name ;
	char *config_file ;
} opts_t; 

int8_t buff[ARTNET_MAX_DMX +1] ;
int length ;

int do_write(artnet_node n, opts_t *ops, int length) {
	int res ;
	
	res = write(ops->fd, buff, length+1);

	if (res < 0){
		/* if you unplug devices from the dongle
		 * it causes a write error, in this case we stop trying to handle 
		 * dmx packets, and set an alarm to poll
		 */
		if(ops->verbose)
			perror("Error writing to device");
		
		res = close(ops->fd) ;
		if(res < 0) 
			perror("close ") ;
		else
			ops->fd = UNCONNECTED ;

		// set alarm
		alarm(ALARM_INT) ;
		
		// don't handle dmx for the moment
		artnet_set_dmx_handler(n, NULL, NULL) ;
		return -1 ;
	}

	if(res > 0 && ops->verbose)
		printf("0x%02hhx, 0x%02hhx, 0x%02hhx, 0x%02hhx length %d\n", buff[1], buff[2], buff[3], buff[4] , length) ;
	
	return 0;
}

/*
 * Called when we have dmx data pending
 */
int dmx_callback(artnet_node n, int port, void *d) {
	int8_t *data ;
	opts_t *ops = (opts_t *) d ;

	data = artnet_read_dmx(n, port, &length) ;
	memset(buff, 0x00, ARTNET_MAX_DMX+1) ;
	memcpy(&buff[1], data, length) ;

	do_write(n,ops, length ) ;
	return 0;
}


/*
 * Attempt to open the dmx device, this will fail if the widget is disconnected
 */
void open_widget(artnet_node n, opts_t *ops) {
	ops->fd = open(ops->dev,O_WRONLY) ;

	if(ops->fd > 0) {
		artnet_set_dmx_handler(n, dmx_callback, ops) ;
		// cancel alarms
		alarm(0);
	} else 
		if(ops->verbose)
			perror("open") ;


}

/*
 * saves the settings to the config file
 */
int save_config(opts_t *ops) {
	FILE *fh ;
	printf("in save config\n") ;
	if ((fh = fopen(ops->config_file, "w")) == NULL ) {
		perror("fopen") ;
		return -1 ;
	}

	fprintf(fh, "# artnet_usb config file\n") ;
	fprintf(fh, "Shortname=%s\n", ops->short_name) ;
	fprintf(fh, "Longname=%s\n", ops->long_name) ;
	fprintf(fh, "Subnet=%i\n", ops->subnet_addr) ;
	fprintf(fh, "Port=%i\n", ops->port_addr);

	fclose(fh) ;

	return 0 ;
}

/*
 * Load the settings from config file
 */
int load_config(opts_t *ops) {
	#define BUF_SIZE 1024
	FILE *fh ;
	char buf[1024], *c ;
	char *key, *data ;
	
	if(ops->config_file == NULL)
		return -1;

	if ((fh = fopen(ops->config_file, "r")) == NULL ) {
		perror("fopen") ;
		return -1 ;
	}

	while (  fgets(buf, 1024, fh) != NULL) {
		if(*buf == '#')
			continue ;

		// strip \n
		for(c = buf ; *c != '\n' ; c++) ;
		*c = '\0' ;

		key = strtok(buf, "=") ;
		data = strtok(NULL, "=") ;

		if(key == NULL || data == NULL)
			continue ;

		if(strcmp(key, "Shortname") == 0) {
			free(ops->short_name) ;
			ops->short_name = strdup(data) ;
		} else if(strcmp(key, "Longname") == 0) {
			free(ops->long_name) ;
			ops->long_name = strdup(data) ;
		} else if(strcmp(key, "Subnet") == 0) {
			ops->subnet_addr = atoi(data) ;
		} else if(strcmp(key, "Port") == 0 ) {
			ops->port_addr = atoi(data) ;
		} 

	}

	fclose(fh) ;
	return 0 ;
}

/*
 * called when to node configuration changes, 
 * we need to save the configuration to a file
 */
int program_handler(artnet_node n, void *d) {
	opts_t *ops = (opts_t*) d ;
	artnet_node_config_t config ;
	
	artnet_get_config(n, &config) ;

	free(ops->short_name) ;
	free(ops->long_name) ;
	
	ops->short_name = strdup(config.short_name) ;
	ops->long_name = strdup(config.long_name) ;
	ops->subnet_addr = config.subnet ;
	ops->port_addr = config.out_ports[0] ;

	save_config(ops) ;
	return 0 ;
}


/*
 * The alarm will interupt the select, making sure
 * that we try and re open the dmx device even if there
 * is a lot of artnet traffic
 *
 */
void sig_alarm(int signo) {
	alarm(ALARM_INT) ;	
}

/*
 * Main function
 *
 */
int run(opts_t *ops) {
	artnet_node node ;
	struct sigaction act, oact ;

	// install alarm handler
	act.sa_handler = sig_alarm ;
	sigemptyset(&act.sa_mask) ;
	act.sa_flags = 0 ;

	if(sigaction(SIGALRM, &act, &oact)  < 0)
		printf("handler not installed\n") ;

	load_config(ops) ;

	// create new artnet node, and set config values
    node = artnet_new(ops->ip_addr, ops->verbose) ; ;

	artnet_set_short_name(node, ops->short_name) ;
	artnet_set_long_name(node, ops->long_name) ;
	artnet_set_node_type(node, ARTNET_NODE) ;

	// set the first port to output dmx data
	artnet_set_port_type(node, 0, ARTNET_ENABLE_OUTPUT, ARTNET_PORT_DMX) ;
	artnet_set_subnet_addr(node, ops->subnet_addr) ;

	// set the universe address of the first port
	artnet_set_port_addr(node, 0, ARTNET_OUTPUT_PORT, ops->port_addr) ;

	// we want to be notified when the node config changes
	artnet_set_program_handler(node, program_handler, (void*) ops) ; 

	alarm(ALARM_INT) ;	
	open_widget(node, ops) ;

	artnet_start(node) ;
	
	while(1) {
		// set a 1 second timeout on the read
		// this way we send a DMX frame every second
		// even if we don't get any ArtNet packets
		artnet_read(node, 1) ;

		// try to open widget if unopen
		if(ops->fd == UNCONNECTED)
			open_widget(node,ops) ;
	
		// we might be connected now
		if(ops->fd != UNCONNECTED) {
			do_write(node, ops, 512) ;
		}
	}

	return 0 ;	
}


/*
 * Set our default options, command line args will overide this
 */
void init_ops(opts_t *ops) {
	ops->verbose = 0;
	ops->dev = strdup(DEFAULT_DEVICE) ;
	ops->ip_addr = NULL ;
	ops->subnet_addr = 0 ;
	ops->port_addr = 0 ;
	ops->persist = 0;
	ops->fd = UNCONNECTED ;

	ops->short_name = strdup(SHORT_NAME) ;
	ops->long_name = strdup(LONG_NAME) ;
	ops->config_file = NULL ;
}

/*
 * Parse command lines options and save in opts_s struct
 *
 */
void parse_args(opts_t *ops, int argc, char *argv[]) {
	// parse args
	int optc, port_addr, subnet_addr ;

	// parse options 
	while ((optc = getopt (argc, argv, "c:s:p:d:a:vz")) != EOF) {
		switch (optc) {
 			case 'a':
				free(ops->ip_addr) ;
				ops->ip_addr = (char *) strdup(optarg) ;
		        break;
 			case 'c':
				free(ops->config_file) ;
				ops->config_file = (char *) strdup(optarg) ;
		        break;
			case 'v':
				ops->verbose = 1 ;
		        break;
			case 'd':
				free(ops->dev) ;
				ops->dev = (char *) strdup(optarg) ;
		        break;
			case 's':
				subnet_addr = atoi(optarg) ;

				if(subnet_addr < 0 && subnet_addr > 15) 
					printf("Subnet address must be between 0 and 15\n") ;
				else
					ops->subnet_addr = subnet_addr ;
				break ;
			case 'p':
				port_addr = atoi(optarg) ;

				if(port_addr < 0 && port_addr > 15)
					printf("Port address must be between 0 and 15\n") ;
				else 
					ops->port_addr = port_addr ;
				break ;
			case 'z':
				ops->persist = 1 ;
      		default:
				break;
    	}
	}
}


int main(int argc, char *argv[]) {	
	int statloc, pid ;
	opts_t ops ;
	
	init_ops(&ops) ;
	parse_args(&ops, argc, argv) ;

	// we don't load ops here
	// else if we restart and the options have been remote 
	// programed we'll have to old ones

	if(ops.persist) {
		while(1) {
			if( (pid = fork() ) == 0) {
				// child ;
				run(&ops) ;
				exit(1) ;
			} else {
				// parent
				pid = wait(&statloc) ;
				printf("Child %i terminated, restarting\n", pid) ;
			}
		}
	} else 
		run(&ops) ;

	return 0;
}

